#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <cls.h>
#include <obj.h>
#include <ui.h>
#include <heap_dump.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp-config.h>

enum {
    INFO,
    NUM_COLUMNS
};

static char buf[1024];

typedef struct tc {
    GtkTreeStore* tree;
    cls* clz;
} tc;

static int not_in_stack (jobjectID obj, jobjectID* stack, size_t level) {
    size_t i;
    for (i = 0; i < level; i++) 
	if (obj == stack[i])
	    return 0;
    return 1;
}

static void list_owners (jobjectID obj_id, size_t level, 
			 size_t maxlevel, jobjectID* stack, 
			 GtkTreeStore* tree, GtkTreeIter* parent_node) {
    GtkTreeIter iter;
    object_link* owners;
    if (level >= maxlevel)
	return;
    stack[level] = obj_id;
    owners = get_owners (obj_id);
    while (owners != NULL) {
	obj* parent;
	cls* pc = NULL;
	cls* vc = NULL;
	char* var = NULL;
	parent = get_object (owners->parent);
	if (parent)
	    pc = get_class (obj_get_class_id (parent));
	vc = get_class (owners->clz);
	if (pc && vc) {
	    switch (owners->type) {	    
	    case STATIC_VARIABLE:
		var = vc->statics[owners->variable].field_name;
		snprintf (buf, 1024, _("%p is a static variable (%s) in class %s"),
			  owners->obj, var, vc->name);
		break;
	    case VARIABLE:
		var = vc->instances[owners->variable].field_name;
		snprintf (buf, 1024, _("%p is a variable (%s) in object %p of class %s"),
			  owners->obj, var, owners->parent, pc->name);
		break;
	    case ARRAY:
		var = "[]";
		snprintf (buf, 1024, _("%p is located in an array %p of %s at index %d"),
			  owners->obj, owners->parent, vc->name, owners->index);
		break;
	    }
	} else {
	    snprintf (buf, 1024, _("not able to determine class or parent class, probably GC:ed"));
	}
	
	gtk_tree_store_append (tree, &iter, parent_node);
	gtk_tree_store_set (tree, &iter, INFO, buf, -1);

	if (not_in_stack (owners->parent, stack, level) && 
	    owners->parent != owners->obj &&
	    !owners->type == STATIC_VARIABLE) {
	    list_owners (owners->parent, level + 1, maxlevel, stack, tree, &iter);
	}
	owners = owners->next;
	if (pc == NULL || vc == NULL)
	    owners = NULL;
    }
}

static void find_instance_owners (void* object, void* tz) {
    obj* o = (obj*)object;
    tc* t = (tc*)tz;
    cls* c = t->clz;
    size_t maxlevel = 7;
    jobjectID* save = malloc (sizeof (jobjectID) * maxlevel);
    if (obj_get_class_id (o) == c->class_id) {
	GtkTreeIter iter;
	gtk_tree_model_get_iter_first (GTK_TREE_MODEL (t->tree), &iter);
	list_owners (o->obj_id, 0, maxlevel, save, t->tree, &iter);
    }
}


static void
add_columns (GtkTreeView *treeview)
{
    GtkCellRenderer *renderer;
    renderer = gtk_cell_renderer_text_new ();
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (treeview),
						 -1, _("Class"),
						 renderer, "text",
						 INFO,
						 NULL);
}

void show_instance_owners (hashtab* objects, cls* c) {
    GtkWidget *win;
    GtkWidget *scrolledwindow;
    GtkWidget *tree;    
    GtkTreeIter   iter;
    GtkTreeStore* model;
    tc tz;

    win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (win), _("Instance owners"));
    scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
    gtk_container_add (GTK_CONTAINER (win), scrolledwindow);
    model = gtk_tree_store_new (NUM_COLUMNS, G_TYPE_STRING);

    snprintf (buf, 1024, "%s", cls_get_name (c));
    gtk_tree_store_append (model, &iter, NULL);
    gtk_tree_store_set (model, &iter, INFO, buf, -1); 
    tz.clz = c;
    tz.tree = model;
    hash_for_each_with_arg (find_instance_owners, objects, &tz);

    tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (model));
    gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (tree), TRUE);
    add_columns (GTK_TREE_VIEW (tree));
    gtk_container_add (GTK_CONTAINER (scrolledwindow), tree);
    gtk_widget_set_usize (win, 500, 200);
    gtk_widget_show_all (win);
}

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
