#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <jni.h>
#include <jvmpi.h>
#include <gtk/gtk.h>
#include <jmp-config.h>
#include <hash.h>
#include <jmp.h>
#include <ui.h>
#include <cls.h>
#include <jthread.h>
#include <filterconfig.h>
#include <eventconfig.h>
#include <classlist_menu.h>
#include <methodlist_menu.h>
#include <threads_window.h>
#include <comparators.h>
#include <dumper.h>
#include <heapgraph.h>

/** state tokens. */
static int        init_done = 0;
static int        quit = 0;
static int        is_updating = 1;

/** id of last dump file... */
static int dump_id = -1;

/** Are we dumping on a timer basis? */
static gint  current_timer = 0; /* the timer identifier. */
static long  dumptimer = 0;     /* The countdown value (10)... */
static long  dumptime = 0;      /* Current value (1, 2, 3, 4,....) */

static GtkWidget *statusbar;
static GtkTooltips *tooltips;
static GtkWidget *JMPmain;

/** arrays that hold the methods/objects we show. */
static cls** classlist = NULL;
static jthread** threadslist = NULL;
static method** methodlist = NULL;
static int classes_count = 0;
static int threads_count = 0;
static int methods_count = 0;

/** A class to show total numbers in class info window. */
static cls* total;

/** The window of the current threads..  This is used in the
    timerstack locking functions to decide iff we need to lock. */
static threads_window *threadswin = NULL;

static GtkListStore *object_list;
static int object_list_size = 0;

static GtkListStore *method_list;
static int method_list_size = 0;

/** This is the timeout function called from gtk-main.
 *  Check if it is time to quit.
 */
gint run_updates (gpointer data) {
    if (quit) {
	if (usingUI ())
	    gtk_main_quit ();   /* time to shut down. */
	return FALSE;           /* stop timer. */
    }
    
    if (usingUI ()) {
	updateUI (get_classes (), get_methods ());
    }
    
    if (dumptimer > 0) {
	dumptime++;
	if (dumptime == dumptimer) {
	    run_data_dump ();
	    dumptime = 0;
	}
    }
    
    return TRUE;                /* keep timer */
}

/** Disable closing of window.
 */
static gint delete_event (GtkWidget *widget, GdkEvent  *event, 
			  gpointer   data) {
    return(TRUE);
}

static void destroy (GtkWidget *widget, gpointer data) {
    gtk_main_quit();
}

void run_dump (GtkMenuItem *menuitem, gpointer user_data) {
    run_data_dump ();
}

void run_garbage_collector (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Requesting garbage collection ..."));
    run_GC ();
}

void run_heap_dumper (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Getting heap dump"));
    run_heap_dump ();
    set_status (_("Heap dumped"));
}

void run_reset_counter (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Resetting counters..."));
    reset_counters ();
    set_status (_("Counters reseted"));
}

void run_restore_counter (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Resetoring counters..."));
    restore_counters ();
    set_status (_("Counters restored"));
}

static void freeze_ui () {
    is_updating = !is_updating;
    set_status (is_updating ? 
		_("UI updating continued") : 
		_("UI updating frozen"));
}

static void show_threads () {
    set_status ("showing threads...");
    if (threadswin == NULL)
	threadswin = create_threads_window (get_threads ());
    else 
	update_threads_window (threadswin, get_threads ());
    gtk_widget_show_all (GTK_WIDGET (threadswin->window));
}

/** Originally create by glade, which seems like a nice tool... 
 */
GtkWidget* create_about_dialog (void) {
    GtkWidget *about_dialog;
    GtkWidget *dialog_vbox1;
    GtkWidget *label1;
    GtkWidget *dialog_action_area1;
    GtkWidget *hbuttonbox2;
    guint button6_key;
    GtkWidget *button6;
    GtkAccelGroup *accel_group;

    accel_group = gtk_accel_group_new ();

    about_dialog = gtk_dialog_new ();
    gtk_object_set_data (GTK_OBJECT (about_dialog), "about_dialog", about_dialog);
    gtk_window_set_title (GTK_WINDOW (about_dialog), _("About JMP"));
    gtk_window_set_policy (GTK_WINDOW (about_dialog), TRUE, TRUE, FALSE);

    dialog_vbox1 = GTK_DIALOG (about_dialog)->vbox;
    gtk_object_set_data (GTK_OBJECT (about_dialog), "dialog_vbox1", dialog_vbox1);

    label1 = gtk_label_new (_("JMP - A java profiler by\n"
			      "  Robert 'Ernimril' Olofsson - robo@khelekore.org  \n"
			      "  Bodo Pfelzer \n"
			      "  Marco Walther \n"
			      "  and others"));
    gtk_widget_ref (label1);
    gtk_object_set_data_full (GTK_OBJECT (about_dialog), "label1", label1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_box_pack_start (GTK_BOX (dialog_vbox1), label1, FALSE, FALSE, 0);

    dialog_action_area1 = GTK_DIALOG (about_dialog)->action_area;
    gtk_object_set_data (GTK_OBJECT (about_dialog), "dialog_action_area1", dialog_action_area1);
    gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 10);

    hbuttonbox2 = gtk_hbutton_box_new ();
    gtk_widget_ref (hbuttonbox2);
    gtk_object_set_data_full (GTK_OBJECT (about_dialog), "hbuttonbox2", hbuttonbox2,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbuttonbox2, TRUE, TRUE, 0);

    button6 = gtk_button_new_with_label ("");
    button6_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button6)->child),
					 _("_Ok"));
    gtk_widget_add_accelerator (button6, "clicked", accel_group,
				button6_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button6);
    gtk_object_set_data_full (GTK_OBJECT (about_dialog), "button6", button6,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), button6);
    GTK_WIDGET_SET_FLAGS (button6, GTK_CAN_DEFAULT);

    gtk_signal_connect_object (GTK_OBJECT (button6), "clicked",
			       GTK_SIGNAL_FUNC (gtk_widget_hide),
			       GTK_OBJECT (about_dialog));

    gtk_window_add_accel_group (GTK_WINDOW (about_dialog), accel_group);
    return about_dialog;
}

static void about_jmp (GtkMenuItem *menuitem, gpointer user_data) {
    GtkWidget *about = create_about_dialog ();	
    gtk_widget_show_all (about);
}

/** Originally create by glade, which seems like a nice tool... 
 *  Names of widgets is not totally nice though...
 */
GtkWidget* create_JMP (void) {
    GtkWidget *JMP;
    GtkWidget *vbox;
    GtkWidget *menubar;
    GtkWidget *hbuttonbox;
    GtkWidget *button;
    GtkWidget *heapgraph;
    guint button_key;
    GtkItemFactoryEntry menu_items[] = {
	{ _("/_File"),                  NULL, NULL,                  0, "<Branch>" },
	{ _("/File/_Dump"),             NULL, run_dump,              0, NULL },
	{ _("/File/_Reset counters"),   NULL, run_reset_counter,     0, NULL },
	{ _("/File/Re_store counters"), NULL, run_restore_counter,   0, NULL },
	{ _("/File/System GC"),         NULL, run_garbage_collector, 0, NULL },
	{ _("/File/_Heapdump"),         NULL, run_heap_dumper,       0, NULL },
	{ _("/File/_Freeze UI"),        NULL, freeze_ui,             0, NULL },
	{ _("/File/Threads"),           NULL, show_threads,          0, NULL },
	{ _("/_Options"),               NULL, NULL,                  0, "<Branch>" },
	{ _("/Options/Filter"),         NULL, filter_edit_options,   0, NULL },
	{ _("/Options/Events"),         NULL, event_window,          0, NULL },
	{ _("/_Help"),                  NULL, NULL,                  0, "<LastBranch>" },
	{ _("/_Help/About"),            NULL, about_jmp,             0, NULL },
    };

    GtkItemFactory *item_factory;
    GtkAccelGroup *accel_group;
    gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);


    JMP = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (JMP), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);

    gtk_window_set_title (GTK_WINDOW (JMP), _("Java Memory Profiler - Main"));
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (JMP), vbox);
    
    /* build menues. */
    accel_group = gtk_accel_group_new ();
    item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", 
					 accel_group);
    gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
    gtk_window_add_accel_group (GTK_WINDOW (JMP), accel_group);
    menubar = gtk_item_factory_get_widget (item_factory, "<main>");
    gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, FALSE, 0);
    
    /* build the button bar. */
    hbuttonbox = gtk_hbutton_box_new ();
    gtk_hbutton_box_set_spacing_default (0);
    gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, TRUE, TRUE, 0);

    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("_Dump"));
    gtk_tooltips_set_tip(tooltips, button, _("Dump data to a text file"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (run_dump),
			NULL);


    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("_Reset"));
    gtk_tooltips_set_tip(tooltips, button, _("Reset counters to zero"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (run_reset_counter),
			NULL);

    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("Re_store"));
    gtk_tooltips_set_tip(tooltips, button, _("Restore counters to the standard (full) values"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (run_restore_counter),
			NULL);

    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("System._GC"));
    gtk_tooltips_set_tip(tooltips, button, _("Run the java virtual machines garbage collector"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (run_garbage_collector),
			NULL);

    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("_Heapdump"));
    gtk_tooltips_set_tip(tooltips, button, _("Show the current heap"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (run_heap_dumper),
			NULL);

    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("_Freeze ui"));
    gtk_tooltips_set_tip(tooltips, button, _("Stop updating values in the JMP windows"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (freeze_ui),
			NULL);

    button = gtk_button_new_with_label ("");
    button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child),
					_("Threads"));
    gtk_tooltips_set_tip(tooltips, button, _("Show the current threads"), NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (show_threads),
			NULL);

 
    /* add a heap size graph */
    heapgraph = get_heap_graph ();
    gtk_box_pack_start (GTK_BOX (vbox), heapgraph, TRUE, TRUE, 0);   

    /* add status bar. */
    statusbar = gtk_statusbar_new ();
    gtk_box_pack_start (GTK_BOX (vbox), statusbar, FALSE, FALSE, 0);

    return JMP;
}

/** array of the class compare functions... */
static int ((*cls_comprs[])(const void* v1, const void* v2)) = { cls_compr_name, 
								 cls_compr_instance, 
								 cls_compr_max_instance, 
								 cls_compr_size,
								 cls_compr_instance_gc };

/** our current comparator */
static int (*cls_compr) (const void* v1, const void* v2) = cls_compr_size;

static void cls_column_clicked (GtkWidget *treeviewcolumn, gpointer user_data) {
    int column = (int)user_data;
    if (cls_comprs[column] != NULL) {
	if (cls_compr != cls_comprs[column]) {
	    cls_compr = cls_comprs[column];
	}
    } else {
	fprintf (stdout, "Sort order not yet implemented.\n");
    }
}

static void add_column (GtkWidget* tree, char* label, int pos, 
			gpointer data, 
			void (*cb)(GtkWidget*, gpointer), 
			gint width,  /* set starting width in pixels */
			int align)   /* 0 = left, 0.5 = center, 1 = right */
{
    GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
    GtkTreeViewColumn *column = 
	gtk_tree_view_column_new_with_attributes (label, renderer,
						  "text", pos, NULL);
    /* TODO this cant be the right way to set alignment can it? /robo */
    renderer->xalign = align;
    
    gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
    gtk_tree_view_column_set_resizable (column, TRUE);
    gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
    gtk_tree_view_column_set_min_width (column, width);
    if (cb != NULL)
	g_signal_connect (G_OBJECT (column), "clicked", G_CALLBACK (cb), data);
}

static void build_object_window () {
    GtkWidget* scroller;
    GtkWidget *tree;
    GtkTreeSelection *select;
    
    GtkWidget* jmpwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);
    gtk_window_set_title (GTK_WINDOW (jmpwin), _("Java Memory Profiler - Objects"));
    scroller = gtk_scrolled_window_new (NULL, NULL);
    gtk_container_add (GTK_CONTAINER (jmpwin), scroller);
    
    object_list = gtk_list_store_new (ON_COLUMNS, G_TYPE_STRING, G_TYPE_LONG, 
				      G_TYPE_LONG, G_TYPE_LONG, G_TYPE_LONG, 
				      G_TYPE_POINTER);
    
    tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (object_list));
    add_column (tree, _("Class"), OCLASS_COLUMN, (gpointer)0, cls_column_clicked, 200, 0);
    add_column (tree, _("Instances"), OINSTANCE_COLUMN, (gpointer)1, cls_column_clicked, 80, 1);
    add_column (tree, _("Max instances"), OINSTANCE_MAX_COLUMN, (gpointer)2, cls_column_clicked, 80, 1);
    add_column (tree, _("Size"), OSIZE_COLUMN, (gpointer)3, cls_column_clicked, 80, 1);
    add_column (tree, _("#GC"), OGC_COLUMN, (gpointer)4, cls_column_clicked, 80, 1);
    gtk_tree_view_set_headers_clickable (GTK_TREE_VIEW (tree), TRUE);
    gtk_container_add (GTK_CONTAINER (scroller), tree);    

    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
    g_signal_connect (G_OBJECT (select), "changed",
		      G_CALLBACK (olist_row_changed),
		      object_list);
    gtk_signal_connect (GTK_OBJECT(tree), "button_press_event",
			GTK_SIGNAL_FUNC (olist_button_handler), 
			NULL);

    gtk_widget_set_usize (jmpwin, 550, 200);
    gtk_widget_show_all (jmpwin);	
}

/** array of method compare functions... */
static int ((*method_comprs[])(const void* v1, const void* v2)) = { method_compr_class, 
								    method_compr_class, 
								    method_compr_time, 
								    method_compr_calls,
								    method_compr_hold_time,
								    method_compr_total_time };


/** The current method comparator */
static int (*method_compr) (const void* v1, const void* v2) = method_compr_time;

static void method_column_clicked (GtkWidget *treeviewcolumn, gpointer user_data) {
    int column = (int)user_data;
    if (method_comprs[column] != NULL) {
	if (method_compr != method_comprs[column]) {
	    method_compr = method_comprs[column];
	}
    } else {
	fprintf (stdout, "Sort order not yet implemented.\n");
    }
}

static void build_method_window () {
    GtkWidget* scroller;
    GtkWidget *tree;
    GtkTreeSelection *select;

    GtkWidget* jmpwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);
    gtk_window_set_title (GTK_WINDOW (jmpwin), _("Java Memory Profiler - Methods"));
    scroller = gtk_scrolled_window_new (NULL, NULL);
    gtk_container_add (GTK_CONTAINER (jmpwin), scroller);
    
    method_list = gtk_list_store_new (MN_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, 
				      G_TYPE_STRING, G_TYPE_LONG, G_TYPE_STRING,
				      G_TYPE_STRING, G_TYPE_POINTER);
    
    tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (method_list));
    add_column (tree, _("Class"), MCLASS_COLUMN, (gpointer)0, method_column_clicked, 200, 0);
    add_column (tree, _("Method"), MMETHOD_NAME_COLUMN, (gpointer)1, method_column_clicked, 200, 0);
    add_column (tree, _("secs"), MSECS_COLUMN, (gpointer)2, method_column_clicked, 80, 1);
    add_column (tree, _("calls"), MCALLS_COLUMN, (gpointer)3, method_column_clicked, 80, 1);
    add_column (tree, _("subs sec"), MSUBS_COLUN, (gpointer)4, method_column_clicked, 80, 1);
    add_column (tree, _("total"), MTOTAL_COLUMN, (gpointer)5, method_column_clicked, 80, 1);
    gtk_tree_view_set_headers_clickable (GTK_TREE_VIEW (tree), TRUE);
    gtk_container_add (GTK_CONTAINER (scroller), tree);    

    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
    g_signal_connect (G_OBJECT (select), "changed",
		      G_CALLBACK (mlist_row_changed),
		      method_list);
    gtk_signal_connect (GTK_OBJECT(tree), "button_press_event",
			GTK_SIGNAL_FUNC (mlist_button_handler), 
			NULL);

    gtk_widget_set_usize (jmpwin, 800, 200);
    gtk_widget_show_all (jmpwin);
}

static void setup_object_tracing () {
    if (total == NULL)
	total = cls_new (_("Total"), _("Total"), NULL, 0, 0, NULL, 0, NULL);
    if (object_list == NULL)
	build_object_window ();
}

static void setup_method_tracing () {
    if (method_list == NULL)
	build_method_window ();
}

/** The thread used to run gtk-main. 
 */
void gtkthread (void* data) {
    /* avoid gtk if we are not using any graphical ui */
    if (usingUI ()) {
	int argc = 1;
	char** argv;
	argv = malloc (sizeof (*argv));
	argv[0] = "jmp";
	gtk_init (&argc, &argv);
	if (tracing_objects ()) {
	    setup_object_tracing ();
	}
	if (tracing_methods ()) {
	    setup_method_tracing ();
	}
	
	tooltips = gtk_tooltips_new ();
	JMPmain = create_JMP ();
	gtk_widget_show_all (JMPmain);
	current_timer = gtk_timeout_add (1000, run_updates, NULL);
	init_done = 1;
	set_status (_("Ok"));
	gtk_main ();
    } else {
	while (!quit) {
	    sleep (1);
	    run_updates (NULL);
	}	    
    }    
}

void set_ui_update_interval (int seconds) {
    if (usingUI () && current_timer) {
	gtk_timeout_remove (current_timer);
	current_timer = gtk_timeout_add (seconds * 1000, run_updates, NULL);
    }
}

/** Tell the ui system to shut itself down. */
void quit_ui (void) {
    quit = 1;
    if (total != NULL)
	cls_free (total);
}

/** Count the number of threads that are currently active */
static void count_threads (void* data) {
    threads_count++;
}

/* Add a thread to the array of threads. */
static void add_threads_row (void* data) {
    jthread* t = (jthread*)data;
    /* do tests ? */
    threadslist[threads_count++] = t;
}

static int (*jthread_compr) (const void* v1, const void* v2) = jthread_compr_name;

/** Count the number of classes that have more than zero instances. */
static void count_classes (void* data, cls_usage** cu) {
    cls* c = (cls*)data;
    /* fill in filtered class usage.. */
    if (cls_get_instances (c) != 0 && 
	cls_get_filtered (c)) {
	classes_count++;
	cu[0]->totalAlloced += c->usage.totalAlloced;
	cu[0]->totalInstances += c->usage.totalInstances;
	cu[0]->maximumInstances += c->usage.maximumInstances;
	cu[0]->totalGC += c->usage.totalGC;
    }
    /* fill in heap usage. */
    cu[1]->totalAlloced += c->usage.totalAlloced;
    cu[1]->totalInstances += c->usage.totalInstances;
    cu[1]->maximumInstances += c->usage.maximumInstances;
    cu[1]->totalGC += c->usage.totalGC;    
}

/** Add a row to the array of data. */
static void add_class_row (void* data) {
    cls* c = (cls*)data;    
    if (cls_get_instances (c) != 0 && 
	(c == total || cls_get_filtered (c)))
	classlist[classes_count++] = c;
}

/** Add a row to the gtk CList (update the ui). */
static void add_class_row_to_list (cls* c, int row, GtkTreeIter* iter) {
    int fast = 0;
    if (row < object_list_size) {
	cls* c2;
	gtk_tree_model_get (GTK_TREE_MODEL (object_list), iter, OOBJECT_COLUMN, &c2, -1);
	if (c == c2) {
	    if (!cls_check_modified (c)) {
		gtk_tree_model_iter_next (GTK_TREE_MODEL (object_list), iter);
		return;
	    }
	    fast = 1;
	}
    } else {
	gtk_list_store_append (object_list, iter);
    }
    
    if (fast) {
	gtk_list_store_set (object_list, iter,
			    OINSTANCE_COLUMN, cls_get_instances (c),
			    OINSTANCE_MAX_COLUMN, cls_get_max_instances (c),
			    OSIZE_COLUMN, cls_get_size (c),
			    OGC_COLUMN, cls_get_total_gc (c),
			    -1);
    } else {
	gtk_list_store_set (object_list, iter,
			    OCLASS_COLUMN, cls_get_name (c),
			    OINSTANCE_COLUMN, cls_get_instances (c), 
			    OINSTANCE_MAX_COLUMN, cls_get_max_instances (c),
			    OSIZE_COLUMN, cls_get_size (c),
			    OGC_COLUMN, cls_get_total_gc (c),
			    OOBJECT_COLUMN, c,
			    -1);
    }
    cls_set_modified (c, 0);
    gtk_tree_model_iter_next (GTK_TREE_MODEL (object_list), iter);
}

static void get_time_string (char* buf, size_t len, jlong val) {
    jlong jsec = val / 1000000000ll;
    jlong jnsec = val % 1000000000ll;
    snprintf (buf, len, "%lld.%09lld", jsec, jnsec);
}

/** Add a row to the gtk CList (update the ui). */
static void add_method_row_to_list (method* m, int row, GtkTreeIter* iter) {
    char sec[64];
    char hsec[64];
    char tsec[64];
    methodtime* mt;
    int fast = 0;

    if (row < method_list_size) {
	method* m2;
	gtk_tree_model_get (GTK_TREE_MODEL (method_list), iter, MMETHOD_COLUMN, &m2, -1);
	if (m == m2) {
	    if (!method_check_modified (m)) {
		gtk_tree_model_iter_next (GTK_TREE_MODEL (method_list), iter);
		return;
	    }
	    fast = 1;	    
	}
    } else {
	gtk_list_store_append (method_list, iter);
    }

    mt = method_get_time_used (m);
    get_time_string (sec, 64, mt->tv);
    get_time_string (hsec, 64, mt->tv_hold);
    get_time_string (tsec, 64, mt->tv + mt->tv_hold);
    if (fast) {
	gtk_list_store_set (method_list, iter,
			    MSECS_COLUMN, sec,
			    MCALLS_COLUMN, method_get_calls (m),
			    MSUBS_COLUN, hsec,
			    MTOTAL_COLUMN, tsec,
			    -1);
    } else {
	cls* c = method_get_owner (m);
	gtk_list_store_set (method_list, iter,
			    MCLASS_COLUMN, cls_get_name (c),
			    MMETHOD_NAME_COLUMN, method_get_method_jmpname (m),
			    MSECS_COLUMN, sec,
			    MCALLS_COLUMN, method_get_calls (m),
			    MSUBS_COLUN, hsec,
			    MTOTAL_COLUMN, tsec,
			    MMETHOD_COLUMN, m,
			    -1);
    }
    method_set_modified (m, 0);
    gtk_tree_model_iter_next (GTK_TREE_MODEL (method_list), iter);
}

/** Count the number of methodods that have used any time. */
static void count_methods (void* data) {
    method* m = (method*)data;
    if (!methodtime_is_zero (&m->time_used) &&
	cls_get_filtered (method_get_owner (m)))
	methods_count++;
}

/** Add a row to the array of data. */
static void add_methods_row (void* data) {
    method* m = (method*)data;
    if (!methodtime_is_zero (&m->time_used) && 
	cls_get_filtered (method_get_owner (m)))
	methodlist[methods_count++] = m;
}

static void clear_usage_data (cls_usage **cp) {
    cls_usage* cu = cp[0];
    cls_usage* hu = cp[1];
    
    cu->totalAlloced = 0;
    cu->totalInstances = 0;
    cu->maximumInstances = 0;
    cu->totalGC = 0;
    hu->totalAlloced = 0;
    hu->totalInstances = 0;
    hu->maximumInstances = 0;
    hu->totalGC = 0;
}

/** Update the statistics. */
void updateUI (hashtab* classes, hashtab* methods) {
    int i;
    GtkTreeIter   iter;
    int oldcount = classes_count;    

    if (!init_done)
	return;

    if (!is_updating)
	return;

    if (tracing_objects ()) {
	cls_usage cu, hu;
	cls_usage* cp[2] = {&cu, &hu};
	classes_count = 0;

	clear_usage_data (cp);
	setup_object_tracing ();	
	hash_lock (classes);
	hash_for_each_with_arg ((hash_iter_fa)count_classes, classes, cp); 
	classes_count++;
	add_heap_size_value (hu.totalAlloced, cu.totalAlloced);
	total->usage = cu;
	if (oldcount != classes_count) {
	    classlist = realloc (classlist, classes_count * sizeof (cls*));
	    /* TODO handle NULL here */
	}
	classes_count = 0;
	add_class_row (total);
	hash_for_each ((hash_iter_f)add_class_row, classes);
	hash_unlock (classes);
	qsort (classlist, classes_count, sizeof (cls*), cls_compr);	
	gtk_tree_model_get_iter_first (GTK_TREE_MODEL (object_list), &iter);
	for (i = 0; i < classes_count; i++) 
	    add_class_row_to_list (classlist[i], i, &iter);
	if (classes_count < object_list_size) {
	    gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (object_list), 
					   &iter, NULL, classes_count);
	    for (i = object_list_size - 1; i >= classes_count; i--) {
		/* this updates iter to point to next valid iter.. */
		gtk_list_store_remove (object_list, &iter);  
	    }
	}
	object_list_size = classes_count;
    }

    if (tracing_methods ()) {
	setup_method_tracing ();
	/* update the method list window.. */
	oldcount = methods_count;
	methods_count = 0;

	hash_lock (methods);
	hash_for_each ((hash_iter_f)count_methods, methods);
	if (oldcount != methods_count) {
	    methodlist = realloc (methodlist, 
				  methods_count * sizeof (method*));
	    /* TODO handle NULL here.. */
	}
	methods_count = 0;
	hash_for_each ((hash_iter_f)add_methods_row, methods);
	hash_unlock (methods);
	qsort (methodlist, methods_count, sizeof (method*), method_compr);

	gtk_tree_model_get_iter_first (GTK_TREE_MODEL (method_list), &iter);
	for (i = 0; i < methods_count; i++) 
	    add_method_row_to_list (methodlist[i], i, &iter);
	if (methods_count < method_list_size) {
	    gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (method_list), 
					   &iter, NULL, methods_count);
	    for (i = method_list_size - 1; i >= methods_count; i--) {
		/* this updates iter to point to next valid iter.. */
		gtk_list_store_remove (method_list, &iter);  
	    }
	}
	method_list_size = methods_count;
    }    
	
    if (threadswin != NULL && GTK_WIDGET_VISIBLE (threadswin->window))
	update_threads_window (threadswin, get_threads ());
}

/** Return the id of the dump file. */
int dump_data (char* dumpfile, hashtab* classes, hashtab* methods, hashtab* threads) {
    FILE* f;
    int i;
    int oldcount = classes_count;
    char filename[128];
    char status[128];
    cls_usage cu, hu;
    cls_usage* cp[2] = {&cu, &hu};
    snprintf (filename, 128, "%s-%d.txt", 
	      (dumpfile == NULL ? "jmp_dump" : dumpfile), ++dump_id);    
    snprintf (status, 128, _("dumped to: %s"), filename);
    set_status (status);
    f = fopen (filename, "w");
    if (f == NULL)
	return -1;

    fprintf (f, "Threads\n");
    fprintf (f, "thread\tgroup\tparent\tclass\tmethod\n");
    fprintf (f, "--------------------------------------------------------------\n");
    oldcount = threads_count;
    threads_count = 0;
    hash_for_each ((hash_iter_f)count_threads, threads);
    if (oldcount != threads_count) 
	threadslist = realloc (threadslist, threads_count * sizeof(jthread*));
    threads_count = 0;
    hash_for_each ((hash_iter_f)add_threads_row, threads);
    qsort(threadslist, threads_count, sizeof(jthread*), jthread_compr);
    for (i = 0; i < threads_count; i++){
	dump_jthread_row (threadslist[i], f);
    }

    fprintf (f, "\n\n\nClassdump\n");
    fprintf (f, "class_name\t#instaces\tmax #instances\tsize\t#GC\n");
    fprintf (f, "--------------------------------------------------------------\n");
    classes_count = 0;

    clear_usage_data (cp);
    hash_for_each_with_arg ((hash_iter_fa)count_classes, classes, cp); 
    if (oldcount != classes_count) 
	classlist = realloc (classlist, classes_count * sizeof (cls*));
    classes_count = 0;
    hash_for_each ((hash_iter_f)add_class_row, classes); 
    qsort (classlist, classes_count, sizeof (cls*), cls_compr);
    for (i = 0; i < classes_count; i++) 
	dump_class_row (classlist[i], f);

    fprintf (f, "\n\n\nMethods\n");
    fprintf (f, "class name\tmethod\tsec\tcalls\tsubsec\n");
    fprintf (f, "--------------------------------------------------------------\n");
    oldcount = methods_count;    
    methods_count = 0;
    hash_for_each ((hash_iter_f)count_methods, methods);
    if (oldcount != methods_count) 
	methodlist = realloc (methodlist, methods_count * sizeof (method*));
    methods_count = 0;
    hash_for_each ((hash_iter_f)add_methods_row, methods);
    qsort (methodlist, methods_count, sizeof (method*), method_compr);
    for (i = 0; i < methods_count; i++) 
	dump_method_row (methodlist[i], f);

    fflush (f);
    fclose (f);
    return dump_id;
}

/** Set the dump timer... */
void set_dump_timer (int dt) {
    dumptimer = dt;
}

/** Get the current dump timer. */
int get_dump_timer () {
    return dumptimer;
}

/** Set the status text... */
void set_status (char* text) {
    if (statusbar) {
	guint context_id = 1;        
	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), context_id);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id, text);
    }
}

/** Check if we have any ui-handling to do. */
int events_pending (void) {
    return gtk_events_pending ();
}

/** Update the ui-toolkit (gtk_main_iteration ()) */
int ui_iteration (void) {
    return gtk_main_iteration ();
}

/** check if a threads window is showing. This info is needed to 
 *  decide if we should lock the timer stacks.
 */
int is_threads_window_showing () {
    return threadswin != NULL && 
	GTK_WIDGET_VISIBLE (threadswin->window);
}

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
